import ij.IJ;
import ij.ImagePlus;
import ij.gui.GenericDialog;
import ij.plugin.PlugIn;
import imagescience.image.Image;
import imagescience.transform.Affine;
import java.awt.FileDialog;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.StringTokenizer;
import java.util.Vector;

public class TJ_Affine implements PlugIn, KeyListener {

    private static String mfile = "<Press F1 to open a file dialog>";

    private static final String[] schemes = {
	"nearest neighbor",
	"linear",
	"cubic convolution",
	"cubic B-spline",
	"cubic O-MOMS",
	"quintic B-spline"
    };
    private static int scheme = 1;

    private static String bgvalue = "0.0";

    private static boolean adjust = true;
    private static boolean antialias = false;

    public void run(String arg) {

	if (!TJ.libcheck()) return;
	final ImagePlus imp = TJ.imageplus();
	if (imp == null) return;

	TJ.debug(TJ.name()+" "+TJ.version()+": Affine");

	GenericDialog gd = new GenericDialog(TJ.name()+": Affine");
	gd.addStringField("Matrix file:",mfile,30);
	final Vector tfs = gd.getStringFields();
	((TextField)(tfs.get(0))).addKeyListener(this);
	gd.addPanel(new Panel(),GridBagConstraints.WEST,new Insets(0,0,0,0));
	gd.addChoice("Interpolation scheme:",schemes,schemes[scheme]);
	gd.addStringField("Background value:",bgvalue);
	gd.addCheckbox(" Adjust size to fit result",adjust);
	gd.addCheckbox(" Anti-alias borders",antialias);
	gd.showDialog();

	if (gd.wasCanceled()) return;

	mfile = gd.getNextString();
	scheme = gd.getNextChoiceIndex();
	bgvalue = gd.getNextString();
	adjust = gd.getNextBoolean();
	antialias = gd.getNextBoolean();

	(new TJAffine()).run(imp,mfile,scheme,bgvalue,adjust,antialias);
    }

    public void keyPressed(KeyEvent e) {

	if ((e.getKeyCode() == KeyEvent.VK_F1) && (e.getSource() instanceof TextField)) {
	    final FileDialog fdg = new FileDialog(IJ.getInstance(),TJ.name()+": Matrix",FileDialog.LOAD);
	    fdg.setFile(""); fdg.setVisible(true);
	    final String dir = fdg.getDirectory();
	    final String file = fdg.getFile();
	    fdg.dispose();
	    if (dir != null && file != null) {
		String path = dir + file;
		if (File.separator.equals("\\")) path = path.replace('\\','/');
		((TextField)e.getSource()).setText(path);
	    }
	}
    }

    public void keyReleased(KeyEvent e) { }

    public void keyTyped(KeyEvent e) { }

}

class TJAffine { // Separate class to make the plugin class independent of imagescience.jar

    void run(final ImagePlus imp,
	     final String mfile,
	     final int scheme,
	     final String bgvalue,
	     final boolean adjust,
	     final boolean antialias) {

	try {
	    final double[][] matrix = load(mfile);
	    final Image img = Image.wrap(imp);
	    final Affine affiner = new Affine();
	    affiner.debug(TJ_Options.debug);
	    double bg;
	    try { bg = Double.parseDouble(bgvalue); }
	    catch (Exception e) { throw new IllegalArgumentException("Invalid background value"); }
	    affiner.background(bg);
	    int ischeme = Affine.NEAREST;
	    switch (scheme) {
	    case 0: ischeme = Affine.NEAREST; break;
	    case 1: ischeme = Affine.LINEAR; break;
	    case 2: ischeme = Affine.CUBIC; break;
	    case 3: ischeme = Affine.BSPLINE3; break;
	    case 4: ischeme = Affine.OMOMS3; break;
	    case 5: ischeme = Affine.BSPLINE5; break;
	    }
	    final Image newimg = affiner.run(img,matrix,ischeme,adjust,antialias);
	    TJ.show(newimg,imp);

	} catch (OutOfMemoryError e) {
	    TJ.error("Not enough memory for this operation");

	} catch (UnknownError e) {
	    TJ.error("Could not create output image for some reason.\nPossibly there is not enough free memory");

	} catch (IllegalArgumentException e) {
	    TJ.error(e.getMessage());

	} catch (Throwable e) {
	    TJ.error("An unidentified error occurred while running the plugin");

	}
    }

    private double[][] load(final String matrixfile) {

	// Read lines:
	final Vector lines = new Vector(10,10);
	String line = null;
	try {
	    final BufferedReader br = new BufferedReader(new FileReader(matrixfile));
	    line = br.readLine();
	    while (line != null) {
		lines.add(line);
		line = br.readLine();
	    }
	    br.close();
	} catch (FileNotFoundException e) {
	    throw new IllegalArgumentException("Unable to find "+matrixfile);
	} catch (Exception e) {
	    throw new IllegalArgumentException("Error reading from "+matrixfile);
	}

	// Convert lines:
	final int nr = lines.size();
	if (nr != 4) throw new IllegalArgumentException("File "+matrixfile+" does not contain a 4x4 matrix");
	String delim = "\t";
	line = (String)lines.get(0);
	if (line.indexOf(",") >= 0) delim = ",";
	else if (line.indexOf(" ") >= 0) delim = " ";
	final double[][] matrix = new double[4][4];
	for (int r=0; r<nr; ++r) {
	    line = (String)lines.get(r);
	    final StringTokenizer st = new StringTokenizer(line,delim);
	    final int nc = st.countTokens();
	    if (nc != 4) throw new IllegalArgumentException("File "+matrixfile+" does not contain a 4x4 matrix");
	    for (int c=0; c<nc; ++c) {
		try { matrix[r][c] = Double.parseDouble(st.nextToken()); }
		catch (Exception e) { throw new IllegalArgumentException("Error reading element ("+r+","+c+") in "+matrixfile); }
	    }
	}

	return matrix;
    }

}
